FracResource

Struct FracResource 

pub struct FracResource<T, const TOTAL: u64> { /* private fields */ }
A struct that stores and dispatches Frac<T>. Unlike Frac, it provides an empty state.

Implementations

Source

impl<T, const TOTAL: u64> FracResource<T, TOTAL>

pub closed spec fn type_inv(self) -> bool

pub closed spec fn storage(self) -> Frac<T, TOTAL>

Returns the Frac<T,TOTAL> stored in this FracResource.

pub open spec fn wf(self) -> bool

{
    &&& TOTAL > 0
    &&& 0 <= self.frac() <= TOTAL
    &&& self.type_inv()

}

Type invariant.

pub open spec fn is_empty(self) -> bool

{ self.frac() == 0 }

Whether this FracResource is empty, i.e., has no fraction.

pub open spec fn not_empty(self) -> bool

{ !self.is_empty() }

Whether the fraction stored in this FracResource is less than TOTAL.

pub open spec fn is_full(self) -> bool

{ self.frac() == TOTAL }

Whether this FracResource has the full fraction, i.e., TOTAL.

pub closed spec fn resource(self) -> T

Returns the value of type T stored in this FracResource.

pub open spec fn view(self) -> T

{ self.resource() }

Returns the value of type T stored in this FracResource. It is an alias of Self::resource.

pub closed spec fn frac(self) -> int

The fractions stored in this FracResource.

pub closed spec fn id(self) -> Loc

Returns the unique identifier.

pub proof fn arbitrary() -> tracked res : Self

requires
TOTAL > 0,

Create an arbitrary FracResource. Useful as a placeholder.

pub proof fn alloc(tracked value: T) -> tracked res : Self

requires
TOTAL > 0,
ensures
res.not_empty(),
res.is_full(),
res@ == value,
res.wf(),

Allocates a new FracResource with the given tracked object.

pub proof fn alloc_from_empty(tracked empty: Empty<T, TOTAL>, tracked value: T) -> tracked res : Self

requires
TOTAL > 0,
ensures
res.is_full(),
res.id() == empty.id(),
res.view() == value,
res.wf(),

Allocates a new FracResource from an Empty<T,TOTAL> with the given tracked object.

pub proof fn split_one(tracked &mut self) -> tracked res : Frac<T, TOTAL>

requires
old(self).not_empty(),
ensures
self.id() == old(self).id(),
self.frac() + 1 == old(self).frac(),
old(self).frac() > 1 ==> self@ == old(self)@,
res.frac() == 1,
res.id() == self.id(),
res.resource() == old(self)@,
old(self).frac() == 1 ==> self.is_empty(),
self.wf(),

Splits a Frac with fraction 1.

pub proof fn split(tracked &mut self, n: int) -> tracked res : Frac<T, TOTAL>

requires
1 <= n <= old(self).frac(),
ensures
self.id() == old(self).id(),
self.frac() + n == old(self).frac(),
old(self).frac() > n ==> self@ == old(self)@,
res.frac() == n,
res.id() == self.id(),
res.resource() == old(self)@,
old(self).frac() == n ==> self.is_empty(),
self.wf(),

Splits a Frac with the given fraction.

pub proof fn combine(tracked &mut self, tracked other: Frac<T, TOTAL>)

requires
old(self).id() == other.id(),
ensures
old(self).frac() + other.frac() > TOTAL ==> false,
old(self).frac() + other.frac() <= TOTAL
    ==> {
        &&& self.id() == old(self).id()
        &&& self.resource() == other.resource()
        &&& self.frac() == old(self).frac() + other.frac()
        &&& self.wf()
        &&& old(self).frac() > 0 ==> self@ == old(self)@

    },

Combines a Frac.

pub proof fn validate(tracked &self)

ensures
self.wf(),

FracResource satisfies the type invariant.

pub proof fn validate_full(tracked &self)

requires
self.is_full(),
ensures
self.not_empty(),
self.frac() == TOTAL,
self.wf(),
pub proof fn validate_with_frac(tracked &self, tracked frac: &Frac<T, TOTAL>)

requires
self.id() == frac.id(),
self.frac() > 0,
ensures
self.resource() == frac.resource(),
pub proof fn take_resource(tracked self) -> tracked (T, Empty<T, TOTAL>)

requires
self.is_full(),
ensures
res == self@,
empty.id() == self.id(),

Consumes the token and takes out the resource.

pub proof fn update(tracked &mut self, tracked value: T) -> tracked res : T

requires
old(self).is_full(),
ensures
self.is_full(),
res == old(self)@,
self.id() == old(self).id(),
self.wf(),

Updates the resource stored in this FracResource and retunrs the old resource if it exists. The fraction must be full before the update.